IDI_ICON1 ICON DISCARDABLE "resources/icons/mask.ico"

#include <windows.h>             // needed for VERSIONINFO
#include "../../libi2pd/version.h"

VS_VERSION_INFO VERSIONINFO
FILEVERSION     I2PD_VERSION_MAJOR,I2PD_VERSION_MINOR,I2PD_VERSION_MICRO,I2PD_VERSION_PATCH
PRODUCTVERSION  I2P_VERSION_MAJOR,I2P_VERSION_MINOR,I2P_VERSION_MICRO,I2P_VERSION_PATCH
FILEOS          VOS_NT_WINDOWS32
FILETYPE        VFT_APP
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904E4" // U.S. English - multilingual (hex)
        BEGIN
            VALUE "CompanyName",        "PurpleI2P"
            VALUE "FileDescription",    "I2Pd Qt"
            VALUE "FileVersion",        I2PD_VERSION
            VALUE "InternalName",       "i2pd-qt"
            VALUE "LegalCopyright",     "Copyright (C) 2013-2018, The PurpleI2P Project"
            VALUE "LegalTrademarks1",   "Distributed under the BSD 3-Clause software license, see the accompanying file COPYING or https://opensource.org/licenses/BSD-3-Clause."
            VALUE "OriginalFilename",   "i2pd_qt.exe"
            VALUE "ProductName",        "i2pd-qt"
            VALUE "ProductVersion",     I2P_VERSION
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x0, 1252 // language neutral - multilingual (decimal)
    END
END
